//----------------------------------------------------------------------------
// oCetB^
// Copyright (c) 2008 N099
//
// QURL: Bilateral Filtering for Gray and Color Images
// http://homepages.inf.ed.ac.uk/rbf/CVonline/LOCAL_COPIES/MANDUCHI1/Bilateral_Filtering.html
//----------------------------------------------------------------------------

//----------------------------------------------------------------------------
// ^[Qbg̃eNX`
//----------------------------------------------------------------------------
Texture2D txSrc;

sampler samSrc =
sampler_state
{
    Texture = <txSrc>;
    AddressU = MIRROR;
    AddressV = MIRROR;
    MinFilter = POINT;
    MagFilter = POINT;
};

Texture2D txRGB;

sampler samRGB =
sampler_state
{
    Texture = <txRGB>;
    AddressU = MIRROR;
    AddressV = MIRROR;
    MinFilter = POINT;
    MagFilter = POINT;
};

//----------------------------------------------------------------------------
// tB^W
//----------------------------------------------------------------------------
#define SZ_WINDOW 3 //ӂ2*SZ_WINDOW+1
float pixGain;
float pixDist;   //Fd
float posDist;   //ʒud
float2 posPitch;  //ItZbgsb`

//----------------------------------------------------------------------------
// RGB <-> YC (CCIR 601-1:derived from TIFF 6.0 section 21, 1992/6/3)
//----------------------------------------------------------------------------
const float3x3 matYC2RGB = 
{ 1.0f,      0.0f,  1.40200f,
  1.0f, -0.34414f, -0.71414f,
  1.0f,  1.77200f,      0.0f
};
  
const float3x3 matRGB2YC = 
{  0.29900f,   0.5870f,    0.114f,
  -0.16874f, -0.33126f,  0.50000f,
   0.50000f, -0.41869f, -0.08131f
};

//----------------------------------------------------------------------------
// KEX֐̑ߎ(x2==x^2)
//----------------------------------------------------------------------------
float gauss2(float x2)
{
	float x,x3;

	if(x2 <= 0.0f){
		return 1.0f;
	}else{
		x2 = saturate(0.6f * 0.6f * x2);
		x3 = x2 * x2 * rsqrt(x2);
		return 2.0f * x3 - 3.0f * x2 + 1.0f;
	}
}

//-----------------------------------------------------------------------------
// Pixel Shader: ToRGBPS
// Desc: RGBɕϊ(int16)
//-----------------------------------------------------------------------------
float4 ToRGBPS(float2 Tex : TEXCOORD0) : COLOR0
{
	float3 ycc = tex2D(samSrc, Tex).rgb;
	return float4(mul(matYC2RGB, ycc), 0.0f);
}

//-----------------------------------------------------------------------------
// Pixel Shader: ToRGBUPS
// Desc: RGBɕϊ(uint16)
//-----------------------------------------------------------------------------
float4 ToRGBUPS(float2 Tex : TEXCOORD0) : COLOR0
{
	//uint16Ȃ̂ŃItZbgĕϊĂēxItZbgZ
	float3 ycc = tex2D(samSrc, Tex).rgb - 32768.0f/65535.0f;
	float3 rgb = mul(matYC2RGB, ycc) + 32768.0f/65535.0f;
	return float4(rgb, 0.0f);
}

//-----------------------------------------------------------------------------
// Pixel Shader: BilateralFilteringPS
// Desc: oCetB^(int16)
//-----------------------------------------------------------------------------
float4 BilateralFilteringPS(float2 Tex : TEXCOORD0, uniform bool last) : COLOR0
{
	float3 rgb,ycc;
	float3 sum = { 0.0f, 0.0f, 0.0f };
	float wsum = 0.0f;
	int i,j;
	
	//Sf
	rgb = tex2D(samRGB, Tex).rgb;
	
	for(i=-SZ_WINDOW; i<=SZ_WINDOW; ++i){
		for(j=-SZ_WINDOW; j<=SZ_WINDOW; ++j){
			float2 pos = { j, i };
			float4 tpos = float4(Tex + posPitch * pos, 0.0f, 0.0f);
			
			float3 col = tex2Dlod(samRGB, tpos).rgb;
			float3 cdif = rgb - col;
			float dist = gauss2(posDist * dot(pos, pos));
			float cdis = gauss2(pixDist * dot(cdif, cdif));
			
			float w = dist * cdis;
			wsum += w;
			sum += col * w;
		}
	}
	
	//K
	sum /= wsum;
	
	//ŏIo
	if(last){
		ycc = tex2D(samSrc, Tex).rgb;
		ycc = lerp(ycc, mul(matRGB2YC, sum), pixGain);
		return float4(ycc, 0.0f);
	}else{
		return float4(sum, 0.0f);
	}
}

//-----------------------------------------------------------------------------
// Pixel Shader: BilateralFilteringUPS
// Desc: oCetB^(uint16)
//-----------------------------------------------------------------------------
float4 BilateralFilteringUPS(float2 Tex : TEXCOORD0, uniform bool last) : COLOR0
{
	float3 rgb,ycc;
	float3 sum = { 0.0f, 0.0f, 0.0f };
	float wsum = 0.0f;
	int i,j;
	
	//Sf
	rgb = tex2D(samRGB, Tex).rgb;
	
	for(i=-SZ_WINDOW; i<=SZ_WINDOW; ++i){
		for(j=-SZ_WINDOW; j<=SZ_WINDOW; ++j){
			float2 pos = { j, i };
			float4 tpos = float4(Tex + posPitch * pos, 0.0f, 0.0f);
			
			float3 col = tex2Dlod(samRGB, tpos).rgb;
			float3 cdif = 2.0f * (rgb - col); //int16Ƃ̃XP[␳
			float dist = gauss2(posDist * dot(pos, pos));
			float cdis = gauss2(pixDist * dot(cdif, cdif));
			
			float w = dist * cdis;
			wsum += w;
			sum += col * w;
		}
	}
	
	//K
	sum /= wsum;
	
	//ŏIo
	if(last){
		ycc = tex2D(samSrc, Tex).rgb;
		rgb = sum - 32768.0f/65535.0f;
		rgb = mul(matRGB2YC, rgb) + 32768.0f/65535.0f;
		ycc = lerp(ycc, rgb, pixGain);
		return float4(ycc, 0.0f);
	}else{
		return float4(sum, 0.0f);
	}
}

//-----------------------------------------------------------------------------
// Technique: BilateralFiltering
// Desc: oCetB^
//-----------------------------------------------------------------------------
technique BilateralFiltering
{
	pass p0
	{
		VertexShader = null;
		PixelShader = compile ps_3_0 ToRGBPS();
		ZEnable = false;
	}

	pass p1
	{
		VertexShader = null;
		PixelShader = compile ps_3_0 BilateralFilteringPS(false);
		ZEnable = false;
	}

	pass p2
	{
		VertexShader = null;
		PixelShader = compile ps_3_0 BilateralFilteringPS(true);
		ZEnable = false;
	}
}

//-----------------------------------------------------------------------------
// Technique: BilateralFilteringU
// Desc: oCetB^(uint16)
//-----------------------------------------------------------------------------
technique BilateralFilteringU
{
	pass p0
	{
		VertexShader = null;
		PixelShader = compile ps_3_0 ToRGBUPS();
		ZEnable = false;
	}

	pass p1
	{
		VertexShader = null;
		PixelShader = compile ps_3_0 BilateralFilteringUPS(false);
		ZEnable = false;
	}

	pass p2
	{
		VertexShader = null;
		PixelShader = compile ps_3_0 BilateralFilteringUPS(true);
		ZEnable = false;
	}
}
