//----------------------------------------------------------------------------------
// aviutl  GPUoCetB^
// Copyright (c) 2008  N099
//----------------------------------------------------------------------------------
#include <math.h>
#include <float.h>
#include <malloc.h>
#include <mmintrin.h>
#include <new>

#include <d3d9.h>
#include <d3dx9.h>
#include <d3dx9math.h>
#include <dxerr.h>

#include "resource.h"
#include "filter.h"

#include "prof.h"
#include "thrproc.h"
#include "gpufilter.h"

//---------------------------------------------------------------------
// struct
//---------------------------------------------------------------------
struct PPVERT
{
    float x, y, z, rhw;
    float tu, tv;       // Texcoord

    const static D3DVERTEXELEMENT9 Decl[3];
};

const D3DVERTEXELEMENT9 PPVERT::Decl[3] =
{
	{ 0, 0,  D3DDECLTYPE_FLOAT4, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITIONT, 0 },
	{ 0, 16, D3DDECLTYPE_FLOAT2, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD,  0 },
	D3DDECL_END()
};

struct MTCP_PARAM
{
	int h, w;
	int sPitch, ycPitch;
	BYTE *pSurf;
	PIXEL_YC *pYC;
};

//---------------------------------------------------------------------
// internal function prototypes
//---------------------------------------------------------------------
static inline void yc_i16_mmx(BYTE*, const PIXEL_YC*, unsigned int);
static inline void i16_yc_mmx(PIXEL_YC*, const BYTE*, unsigned int);
static inline void yc_u16_mmx(BYTE*, const PIXEL_YC*, unsigned int);
static inline void u16_yc_mmx(PIXEL_YC*, const BYTE*, unsigned int);
static void __cdecl cp_yc_i16_mt(int, int, void*, void*);
static void __cdecl cp_i16_yc_mt(int, int, void*, void*);
static void __cdecl cp_yc_u16_mt(int, int, void*, void*);
static void __cdecl cp_u16_yc_mt(int, int, void*, void*);

static volatile long mt_cnt;

//---------------------------------------------------------------------
// public functions
//---------------------------------------------------------------------

//RXgN^
GpuFilter::GpuFilter() :
	m_pD3D(NULL), m_pd3dDevice(NULL),
	m_pSrcTex(NULL), m_pSrfSys(NULL),
	m_pVert(NULL), m_pVertDec(NULL),
	m_pEffect(NULL), m_pQuery(NULL),
	m_tproc(NULL)
{
	m_width = m_height = m_wMax = -1;
	m_nPass = 0;

	m_pixFormat = U16;
	ZeroMemory(m_pRtTex, sizeof(m_pRtTex));
	ZeroMemory(m_pixFormatAvailable, sizeof(m_pixFormatAvailable));

	ZeroMemory(m_hParam, sizeof(m_hParam));
	ZeroMemory(m_hTec, sizeof(m_hTec));
	m_hTex = NULL;
}

//fXgN^
GpuFilter::~GpuFilter()
{
	Release();
	delete m_tproc;
}

//IuWFNg̏
bool GpuFilter::Init(FILTER *fp)
{
	HRESULT hr;

	//Ŝ
	Release();

	//init D3D    
	m_pD3D = Direct3DCreate9(D3D_SDK_VERSION);
	if(NULL == m_pD3D) {
		ShowMessageBox(__FILE__, (DWORD)__LINE__, E_FAIL, TEXT("Direct3DCreate9()s܂"));
		return false;
	}

	//create device
	D3DPRESENT_PARAMETERS d3dpp; 

	ZeroMemory(&d3dpp, sizeof(d3dpp));
	d3dpp.Windowed = TRUE;
	d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
	d3dpp.BackBufferFormat = D3DFMT_UNKNOWN;

	hr = m_pD3D->CreateDevice(
		D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, fp->hwnd,
		D3DCREATE_SOFTWARE_VERTEXPROCESSING|D3DCREATE_NOWINDOWCHANGES|D3DCREATE_MULTITHREADED, 
		&d3dpp, &m_pd3dDevice );
	if(FAILED(hr)) {
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("D3DfoCX̍쐬Ɏs܂"));
		Release();
		return false;
	}


	//foCX̃`FbN
	if(FAILED(CheckDeviceCap())){
		Release();
		return false;
	}


	// ɑ݂郊\[X쐬

	//Xbhp
	SAFE_DELETE(m_tproc);
	m_tproc = new(std::nothrow) ThrProc;
	if(m_tproc == NULL){
		hr = E_POINTER;
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("Xbh쐬Ɏs܂"));
		Release();
		return false;
	}
	if(m_tproc->Init(AsyncProc, this)){
		hr = E_FAIL;
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("XbhɎs܂"));
		Release();
		return false;
	}

	//GtFNg̃[h
	hr = CreateEffect(MAKEINTRESOURCE(IDR_EFFECT_FILE), fp->dll_hinst);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("GtFNg̃[hɎs܂"));
		Release();
		return false;
	}

	//_錾̍쐬
	SAFE_RELEASE(m_pVertDec);
    hr = m_pd3dDevice->CreateVertexDeclaration(PPVERT::Decl, &m_pVertDec);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("_錾̍쐬Ɏs܂"));
		Release();
		return false;
	}

#ifdef PROFILE_GPU
	//fobOp
    hr = CreateQuery();
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("NG̍쐬Ɏs܂"));
		Release();
		return false;
	}
#endif

	return true;
}

//e탊\[X̊m
bool GpuFilter::AllocResources(int w, int h, int w_max, int fmt)
{
	//̃`FbN
	if(!m_pEffect) return false;

	//sNZTCỸZbgi4096x4096ɐj
	if(w<0 || w>4096 || h<0 || h>4096 || w_max<0 || w_max<w){
		return false;
	}else{
		m_width = w;
		m_height = h;
		m_wMax = w_max;
	}

	//sNZtH[}bg
	SetPixelFmt(static_cast<GpuFilter::PIXEL_FORMAT>(fmt));
	
	//ꉞJĂ
	ReleaseResources();

	//VXeɍ쐬郊\[X
	if(FAILED(CreateSysMemResources())){
		return false;
	}

	//GPUɍ쐬郊\[X
	if(FAILED(CreateDefaultResources())){
		return false;
	}

	//GtFNgɃ\[XZbg
	if(FAILED(SetEffectResources())){
		return false;
	}

	return true;
}

//D3D\[X̊J
bool GpuFilter::Release()
{
	//\[X̊J
	SAFE_RELEASE(m_pQuery);
	SAFE_RELEASE(m_pEffect);
	SAFE_RELEASE(m_pVertDec);
	ReleaseResources();

	//AddRefĂ
	SAFE_RELEASE(m_pd3dDevice);
	SAFE_RELEASE(m_pD3D);

	return true;
}

//e탊\[X̊J
bool GpuFilter::ReleaseResources()
{	
	//\[X̊J
	SAFE_RELEASE(m_pVert);
	SAFE_RELEASE(m_pSrfSys);
	SAFE_RELEASE(m_pSrcTex);
	for(int i=0; i<ARRAYSIZE(m_pRtTex); ++i){
		SAFE_RELEASE(m_pRtTex[i]);
	}

	return true;
}

//tB^p[^̃Zbg
bool GpuFilter::SetParams(int *track, int *check)
{
	HRESULT hr;
	float para;

	//̃`FbN
	if(!m_pEffect) return false;

	_mm_empty();

	//
	if(m_hParam[PARAM_GAIN]){
		para = track[T_GAIN]/256.0f;
		hr = m_pEffect->SetFloat(m_hParam[PARAM_GAIN], para);
		if(FAILED(hr)){
			ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, 
				TEXT("p[^1̃ZbgɎs܂"));
		}
	}
	//F̍
	if(m_hParam[PARAM_SIM]){
		para = track[T_SIM]/128.0f;
		para = para * para * 256.0f;
		hr = m_pEffect->SetFloat(m_hParam[PARAM_SIM], para*para);
		if(FAILED(hr)){
			ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, 
				TEXT("p[^2̃ZbgɎs܂"));
		}
	}
	//͈
	if(m_hParam[PARAM_DOM]){
		para = track[T_DOM]/128.0f;
		hr = m_pEffect->SetFloat(m_hParam[PARAM_DOM], para*para);
		if(FAILED(hr)){
			ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, 
				TEXT("p[^3ZbgɎs܂"));
		}
	}

	//
	m_nPass = track[T_NPAS];
	if(m_nPass < 1 || m_nPass >5) m_nPass = 1;

	return true;
}

//foCX̏ԂeXgAKvȂ烊Zbg
bool GpuFilter::TestDeviceState(bool *need_realloc)
{
	HRESULT hr;

	if(!m_pd3dDevice){
		return false;
	}

	// D3D9 device ̏ԂeXg
	hr = m_pd3dDevice->TestCooperativeLevel();
	switch(hr)
	{
		case D3D_OK:
			break;
		case D3DERR_DEVICELOST:
#ifdef PROFILE_GPU
			ShowMessageBox(__FILE__, (DWORD)__LINE__, hr,
				TEXT("foCXXg܂Bs܂B"));
#endif
			return false;
		case D3DERR_DEVICENOTRESET:
			hr = Reset();
			if(FAILED(hr)){
				ShowMessageBox(__FILE__, (DWORD)__LINE__, hr,
					TEXT("foCXZbgɎs܂"));
				return false;
			}else{
				*need_realloc = true;
			}
			break;
		default :
			ShowMessageBox(__FILE__, (DWORD)__LINE__, hr,
				TEXT("TestCooperativeLevel()G[Ԃ܂"));
			return false;
	}

	return true;
}

//PIXEL_YCf[^VXeT[tFXɃRs[
bool GpuFilter::PutImage(const PIXEL_YC *data, EXEC_MT func_mt)
{
	HRESULT hr;

	getCount(TEXT("TestDev "));

	hr = PutPixelData(m_pSrfSys, data, func_mt);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("T[tFXփRs[ł܂ł"));
		return false;
	}

	return true;
}

//f[^̏s
bool GpuFilter::ProcImage(int empty)
{
	if(!m_pSrfSys || !m_pd3dDevice){
		return false;
	}

	//_OJn
	if(SUCCEEDED(m_pd3dDevice->BeginScene()))
	{
		//f[^̃Rs[
		UpdateSurface(m_pSrfSys, m_pSrcTex);
		m_pEffect->SetTexture(m_hTex, m_pRtTex[0]);

		//KȃeNjbNw
		if(m_pixFormat == U16){
			m_pEffect->SetTechnique(m_hTec[U16]);
		}else{ // (m_pixFormat == S16){
			m_pEffect->SetTechnique(m_hTec[S16]);
		}

		// Set the vertex declaration
		m_pd3dDevice->SetVertexDeclaration(m_pVertDec);

		// Draw the quad
		UINT cPasses;
		m_pEffect->Begin(&cPasses, 0);

		// Convert to RGB
		SetRenderTarget(m_pRtTex[0]);
		m_pEffect->BeginPass(0);
		m_pd3dDevice->SetStreamSource(0, m_pVert, 0, sizeof(PPVERT));
		m_pd3dDevice->DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2);
		m_pEffect->EndPass();
		
		// Proc (Repeat)
		int i;
		for(i=1; i<m_nPass; ++i){
			SetRenderTarget(m_pRtTex[i & 0x1]);
			m_pEffect->BeginPass(1);
			m_pd3dDevice->SetStreamSource(0, m_pVert, 0, sizeof(PPVERT));
			m_pd3dDevice->DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2);
			m_pEffect->EndPass();
			m_pEffect->SetTexture(m_hTex, m_pRtTex[i & 0x1]);
		}

		// Proc + Merge
		SetRenderTarget(m_pRtTex[i & 0x1]);
		m_pEffect->BeginPass(2);
		m_pd3dDevice->SetStreamSource(0, m_pVert, 0, sizeof(PPVERT));
		m_pd3dDevice->DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2);
		m_pEffect->EndPass();

		m_pEffect->End();

		// End scene
		m_pd3dDevice->EndScene();

		// empty the command buffer
		if(empty){
			DWORD n;
			m_pd3dDevice->ValidateDevice(&n);
		}
	}

	return true;
}

//f[^̏siʃXbhŔ񓯊j
bool GpuFilter::ProcImageAsync()
{
	if(!m_pSrfSys || !m_pd3dDevice || !m_tproc){
		return false;
	}
	
	return m_tproc->Start() ? false : true;
}

//PIXEL_YCf[^foCX]
bool GpuFilter::GetImage(PIXEL_YC *data, EXEC_MT func_mt)
{
	HRESULT hr;

	if(!data || !m_tproc){
		return false;
	}

	if(m_tproc->isRunning()){
		//҂
		m_tproc->Wait();
	}else{
		//VXeɓ]
		hr = GetRenderTargetData();
		if(FAILED(hr)){
			ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("RTf[^̃Rs[Ɏs܂"));
			return false;
		}
	}

#ifdef PROFILE_GPU
	getCount(TEXT("GetRT   "));
#endif

	//VXeϊăRs[
	hr = GetPixelData(data, m_pSrfSys, func_mt);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("T[tFXRs[ł܂ł"));
		return false;
	}

	return true;
}

//eNX`tH[}bg̕\
const TCHAR *GpuFilter::GetPixelFormatStr(void) const
{
	switch(m_pixFormat){
		case S16:
			return TEXT("S16");
		case U16:
			return TEXT("U16");
		default:
			return TEXT("Unknown");
	}
}

//GPUł̏I܂ő҂
#ifdef PROFILE_GPU
void GpuFilter::WaitForIdle()
{
	if(m_pQuery){
		m_pQuery->Issue(D3DISSUE_END);
		while(S_FALSE == m_pQuery->GetData(NULL, 0, D3DGETDATA_FLUSH))
			__asm pause ;
	}
}
#endif

//---------------------------------------------------------------------
// private functions
//---------------------------------------------------------------------

//foCX̃T|[g󋵂̃`FbN
HRESULT GpuFilter::CheckDeviceCap()
{
	HRESULT hr;

	//|C^̃`FbN
	if(!m_pd3dDevice || !m_pD3D) return D3DERR_INVALIDCALL;

	//D3DCaps̃`FbN
	D3DCAPS9 caps;
	ZeroMemory(&caps, sizeof(caps));
	hr = m_pd3dDevice->GetDeviceCaps(&caps);
	if(D3D_OK != hr) return hr;

	//eNX`֘ÃeXg
	if(caps.TextureCaps & D3DPTEXTURECAPS_SQUAREONLY){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("񐳕eNX`𗘗pł܂"));
		return D3DERR_NOTAVAILABLE;
	}
	if(!(caps.TextureCaps & D3DPTEXTURECAPS_POW2)){
		; //2̗ݏȊÕTCỸeNX`
	}else if(!(caps.TextureCaps & D3DPTEXTURECAPS_NONPOW2CONDITIONAL)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("2̗ݏTCYȊÕeNX`͗pł܂"));
		return D3DERR_NOTAVAILABLE;
	}

    //RTtH[}bg̃`FbN
    D3DDISPLAYMODE DisplayMode;
    hr = m_pd3dDevice->GetDisplayMode( 0, &DisplayMode );
	if(D3D_OK != hr) return hr;

	bool found = false;
	if(SUCCEEDED(m_pD3D->CheckDeviceFormat(caps.AdapterOrdinal, caps.DeviceType,
		DisplayMode.Format, D3DUSAGE_RENDERTARGET, D3DRTYPE_TEXTURE, D3DFMT_A16B16G16R16)))
	{
		found = true;
		m_pixFormatAvailable[U16] = true;
		SetPixelFmt(U16);
	}else{
#ifdef PROFILE_GPU
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("D3DFMT_A16B16G16R16 pł܂"));
#endif
		m_pixFormatAvailable[U16] = false;
	}

	if(SUCCEEDED(m_pD3D->CheckDeviceFormat(caps.AdapterOrdinal, caps.DeviceType,
		DisplayMode.Format, D3DUSAGE_RENDERTARGET, D3DRTYPE_TEXTURE, D3DFMT_Q16W16V16U16)))
	{
		found = true;
		m_pixFormatAvailable[S16] = true;
		SetPixelFmt(S16);
	}else{
#ifdef PROFILE_GPU
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("D3DFMT_Q16W16V16U16 pł܂"));
#endif
		m_pixFormatAvailable[S16] = false;
	}

	if(!found){   
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("p\ȃsNZtH[}bgЂƂ܂"));
		return D3DERR_NOTAVAILABLE;
	}

	return hr;
}

//_̐錾
HRESULT GpuFilter::CreateVert()
{
	HRESULT hr;

    //foCX̃`FbN
	if(!m_pd3dDevice) return D3DERR_INVALIDCALL;

	//Ŝ
	SAFE_RELEASE(m_pVert);

	const PPVERT Quad[4] =
	{
		{         -0.5f,          -0.5f, 1.0f, 1.0f, 0.0f, 0.0f },
		{ m_width -0.5f,          -0.5f, 1.0f, 1.0f, 1.0f, 0.0f },
		{         -0.5f, m_height -0.5f, 1.0f, 1.0f, 0.0f, 1.0f },
		{ m_width -0.5f, m_height -0.5f, 1.0f, 1.0f, 1.0f, 1.0f }
	};

	hr = m_pd3dDevice->CreateVertexBuffer(sizeof(PPVERT) * 4,
		D3DUSAGE_WRITEONLY, 0, D3DPOOL_DEFAULT, &m_pVert, NULL );

	//obt@ɓ]
	LPVOID pVBData;
	if(SUCCEEDED(m_pVert->Lock(0, 0, &pVBData, D3DLOCK_NOSYSLOCK))){
		CopyMemory(pVBData, Quad, sizeof(Quad));
		m_pVert->Unlock();
	}

	return hr;
}

//GtFNgt@CǂݍŃGtFNg쐬
HRESULT GpuFilter::CreateEffect(const TCHAR *res, HMODULE hm)
{
	const DWORD ShaderFlags = D3DXFX_NOT_CLONEABLE;
	ID3DXBuffer *pCompErr = NULL; //RpC̃G[
	HRESULT hr;

	//foCX̃`FbN
	if(!m_pd3dDevice) return D3DERR_INVALIDCALL;

	//Ŝ
	SAFE_RELEASE(m_pEffect);

	//GtFNgt@C̓ǂݎ
	hr = D3DXCreateEffectFromResource(m_pd3dDevice, hm,
		   res, NULL, NULL, ShaderFlags, NULL, &m_pEffect, &pCompErr);
	if(D3D_OK != hr){
		if(pCompErr){
			//RpCG[\
			char *msg = (char*)pCompErr->GetBufferPointer();
#ifdef UNICODE
			if(msg){
				size_t count = pCompErr->GetBufferSize();
				WCHAR *str = new(std::nothrow) WCHAR[count];
				if(str){
					mbstowcs_s(NULL, str, count, msg, count - 1);
					str[count - 1] = L'\0';
					ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, str);
					delete[] str;
				}
			}
#else
			if(msg){
				ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, msg);
			}
#endif
			pCompErr->Release();
		}
		
	}else{
		//p[^nh̎擾
		m_hParam[PARAM_GAIN] = m_pEffect->GetParameterByName(NULL, "pixGain");
		m_hParam[PARAM_SIM]  = m_pEffect->GetParameterByName(NULL, "pixDist");
		m_hParam[PARAM_DOM]  = m_pEffect->GetParameterByName(NULL, "posDist");

		//eNjbN
		m_hTec[S16] = m_pEffect->GetTechniqueByName("BilateralFiltering");
		m_hTec[U16] = m_pEffect->GetTechniqueByName("BilateralFilteringU");

		//eNX`
		m_hTex = m_pEffect->GetParameterByName(NULL, "txRGB");
	}

	SAFE_RELEASE(pCompErr);

	return hr;
}

//VXeɕێ郊\[X쐬
HRESULT GpuFilter::CreateSysMemResources()
{
	HRESULT hr;

	//VXẽT[tFX
	hr = CreateSurface(&m_pSrfSys, 0, D3DPOOL_SYSTEMMEM);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("T[tFX̍쐬Ɏs܂"));
		return hr;
	}

	return hr;
}

//GPUɑ݂郊\[X쐬FXgɂ͕AKv
HRESULT GpuFilter::CreateDefaultResources(void)
{
	HRESULT hr;

	//_
	hr = CreateVert();
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("_̃ZbgɎs܂"));
		return hr;
	}

	//\[XeNX`
	hr = CreateTexture(&m_pSrcTex, 0, D3DPOOL_DEFAULT);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("eNX`0̍쐬Ɏs܂"));
		return hr;
	}

	//RTeNX`1
	hr = CreateTexture(&m_pRtTex[0], D3DUSAGE_RENDERTARGET, D3DPOOL_DEFAULT);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("eNX`1̍쐬Ɏs܂"));
		return hr;
	}

	//RTeNX`2
	hr = CreateTexture(&m_pRtTex[1], D3DUSAGE_RENDERTARGET, D3DPOOL_DEFAULT);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("eNX`2̍쐬Ɏs܂"));
		return hr;
	}

#ifdef PROFILE_GPU
	//fobOp
    hr = CreateQuery();
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("NG̍쐬Ɏs܂"));
		return hr;
	}
#endif
	return hr;
}

//\[XGtFNgɃZbg
HRESULT GpuFilter::SetEffectResources()
{
	HRESULT hr;

	//foCX̃`FbN
	if(!m_pd3dDevice || !m_pEffect) return D3DERR_INVALIDCALL;

	//eNX`̃Zbg
	hr = m_pEffect->SetTexture("txSrc", m_pSrcTex);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("\[XeNX`̃ZbgɎs܂"));
		return hr;
	}

	hr = m_pEffect->SetTexture("txRGB", m_pRtTex[0]);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("RTeNX`̃ZbgɎs܂"));
		return hr;
	}

	_mm_empty();

	//ItZbgsb`̃Zbg
	D3DXVECTOR4 vtmp(1.0f/m_width, 1.0f/m_height, 0.0f, 0.0f);
	hr = m_pEffect->SetVector("posPitch", &vtmp);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("ItZbg̃ZbgɎs܂"));
		return hr;
	}

	return hr;
}

//foCXXgɃZbgs
HRESULT GpuFilter::Reset(void)
{
	HRESULT hr;

	//foCX̃`FbN
	if(!m_pd3dDevice || !m_pEffect) return D3DERR_INVALIDCALL;

	//GPU\[X̊J
	SAFE_RELEASE(m_pVert);
	SAFE_RELEASE(m_pSrcTex);
	for(int i=0; i<ARRAYSIZE(m_pRtTex); ++i){
		SAFE_RELEASE(m_pRtTex[i]);
	}
#ifdef PROFILE_GPU
	SAFE_RELEASE(m_pQuery);
#endif
	m_pEffect->OnLostDevice();

	//Zbg
	D3DPRESENT_PARAMETERS dpp;

	ZeroMemory(&dpp, sizeof(dpp));
	dpp.Windowed = TRUE;
	dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
	dpp.BackBufferFormat = D3DFMT_UNKNOWN;

	hr = m_pd3dDevice->Reset(&dpp);
	if(FAILED(hr)){
		return hr;
	}

	hr = m_pEffect->OnResetDevice();
	if(FAILED(hr)){
		return hr;
	}

	//\[X̍Ċmۂ͂ł͍sȂ
	return hr;
}

//eNX`쐬
HRESULT GpuFilter::CreateTexture(IDirect3DTexture9 **ppTex, DWORD usage, D3DPOOL pool)
{
	D3DFORMAT fmt = PicFmtToD3DFmt(m_pixFormat);

	//foCX̃`FbN
	if(!m_pd3dDevice) return D3DERR_INVALIDCALL;
	
	//Ŝ
	SAFE_RELEASE(*ppTex);

	//eNX`쐬
	return m_pd3dDevice->CreateTexture(m_width, m_height, 1, usage, fmt, pool, ppTex, NULL);
}

//T[tFX쐬
HRESULT GpuFilter::CreateSurface(IDirect3DSurface9 **ppSrf, DWORD usage, D3DPOOL pool)
{
	D3DFORMAT fmt = PicFmtToD3DFmt(m_pixFormat);

	//foCX̃`FbN
	if(!m_pd3dDevice) return D3DERR_INVALIDCALL;

	//Ŝ
	SAFE_RELEASE(*ppSrf);

	if(usage & D3DUSAGE_RENDERTARGET){
		//_[^[Qbg쐬
		return m_pd3dDevice->CreateRenderTarget(m_width, m_height, fmt, D3DMULTISAMPLE_NONE, 0, FALSE, ppSrf, NULL);
	}else{
		//ItXN[T[tFX쐬
		return m_pd3dDevice->CreateOffscreenPlainSurface(m_width, m_height, fmt, pool, ppSrf, NULL);
	}
}

//UpdateTexture𗘗pԐړ]
HRESULT GpuFilter::PutPixelData(IDirect3DSurface9 *sys, const PIXEL_YC *src, EXEC_MT func_mt)
{
	HRESULT hr;
	D3DLOCKED_RECT lRect;

	if(!m_pd3dDevice || !sys) return D3DERR_INVALIDCALL;

	//VXẽT[tFXbN
	hr = sys->LockRect(&lRect, NULL, D3DLOCK_DISCARD|D3DLOCK_NOSYSLOCK);
	if(hr != D3D_OK) return hr;

	getCount(TEXT("PutLock "));

	//f[^̃Rs[
	MULTI_THREAD_FUNC yc_int16_mt = (m_pixFormat == S16) ? cp_yc_i16_mt : cp_yc_u16_mt;
	MTCP_PARAM para = { m_height, m_width, lRect.Pitch, m_wMax,
		static_cast<BYTE*>(lRect.pBits), const_cast<PIXEL_YC*>(src) };
	if(func_mt){
		InterlockedExchange(&mt_cnt, 0);
		func_mt(yc_int16_mt, &para, NULL);
	}else{
		yc_int16_mt(0, 0, &para, NULL);
	}

	getCount(TEXT("PutCopy "));
	
	//f[^̓]Proc()̍ŏɍs
	return sys->UnlockRect();
}

//DynamicTexture𗘗pړ]
HRESULT GpuFilter::PutPixelDataDT(IDirect3DSurface9 *dst, const PIXEL_YC *src, EXEC_MT func_mt)
{
	HRESULT hr;
	D3DLOCKED_RECT lRect;

	if(!dst) return D3DERR_INVALIDCALL;

	//_Ci~bNeNX`̃T[tFXbN
	hr = dst->LockRect(&lRect, NULL, D3DLOCK_DISCARD|D3DLOCK_NOSYSLOCK);
	if(hr != D3D_OK) return hr;

	getCount(TEXT("PutLock "));

	//f[^̓]
	MULTI_THREAD_FUNC yc_int16_mt = (m_pixFormat == S16) ? cp_yc_i16_mt : cp_yc_u16_mt;
	MTCP_PARAM para = { m_height, m_width, lRect.Pitch, m_wMax,
		static_cast<BYTE*>(lRect.pBits), const_cast<PIXEL_YC*>(src) };
	if(func_mt){
		InterlockedExchange(&mt_cnt, 0);
		func_mt(yc_int16_mt, &para, NULL);
	}else{
		yc_int16_mt(0, 0, &para, NULL);
	}

	getCount(TEXT("PutCopy "));

	return dst->UnlockRect();
}

//GetRenderTargetData𗘗pԐړ]
HRESULT GpuFilter::GetRenderTargetData()
{
	HRESULT hr;
	IDirect3DSurface9 *src = NULL;

	if(!m_pd3dDevice){
		return D3DERR_INVALIDCALL;
	}

	//eNX`T[tFX̎擾
	hr = m_pRtTex[m_nPass & 0x1]->GetSurfaceLevel(0, &src);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("RTeNX`T[tFX擾ł܂ł"));
		return hr;
	}

	//VXeɓ]
	hr = m_pd3dDevice->GetRenderTargetData(src, m_pSrfSys);
	SAFE_RELEASE(src);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("foCXRs[ł܂ł"));
		return hr;
	}

	return hr;
}

//VXẽf[^ϊăRs[
HRESULT GpuFilter::GetPixelData(PIXEL_YC *dst, IDirect3DSurface9 *sys, EXEC_MT func_mt)
{
	HRESULT hr;
	D3DLOCKED_RECT lRect;

	if(!m_pd3dDevice || !sys) return D3DERR_INVALIDCALL;

	//VXẽT[tFXbN
	hr = sys->LockRect(&lRect, NULL, D3DLOCK_READONLY|D3DLOCK_NOSYSLOCK);
	if(hr != D3D_OK) return hr;

	getCount(TEXT("GetLock "));

	//f[^̃Rs[
	MULTI_THREAD_FUNC int16_yc_mt = (m_pixFormat == S16) ? cp_i16_yc_mt : cp_u16_yc_mt;
	MTCP_PARAM para = { m_height, m_width, lRect.Pitch, m_wMax,
		static_cast<BYTE*>(lRect.pBits), dst };
	if(func_mt){
		InterlockedExchange(&mt_cnt, 0);
		func_mt(int16_yc_mt, &para, NULL);
	}else{
		int16_yc_mt(0, 0, &para, NULL);
	}

	getCount(TEXT("GetCopy "));

	return sys->UnlockRect();
}

//G[̕\
HRESULT GpuFilter::ShowMessageBox(const char *file, DWORD line, HRESULT hr, const TCHAR *msg)
{
#ifdef _DEBUG
	return DXTrace(file, line, hr, msg, TRUE);
#else
	MessageBox(NULL, msg, TEXT("GPUoCetB^"), MB_ICONERROR|MB_SYSTEMMODAL);
	return S_OK;
#endif
}

//eNX`p
HRESULT GpuFilter::UpdateSurface(IDirect3DSurface9 *src, IDirect3DTexture9 *dst)
{
	HRESULT hr;
	IDirect3DSurface9 *srf = NULL;

	if(!src || !dst || !m_pd3dDevice){
		return D3DERR_INVALIDCALL;
	}

	//T[tFX̎擾
	hr = dst->GetSurfaceLevel(0, &srf);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr,
			TEXT("eNX`T[tFX擾ł܂ł"));
		return hr;
	}

	hr = m_pd3dDevice->UpdateSurface(src, NULL, srf, NULL);
	srf->Release();

	return hr;
}

//eNX`p
HRESULT GpuFilter::SetRenderTarget(IDirect3DTexture9 *dst)
{
	HRESULT hr;
	IDirect3DSurface9 *srf = NULL;

	if(!dst || !m_pd3dDevice){
		return D3DERR_INVALIDCALL;
	}

	//T[tFX̎擾
	hr = dst->GetSurfaceLevel(0, &srf);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr,
			TEXT("eNX`T[tFX擾ł܂ł"));
		return hr;
	}
	
	//RenderTarget̐ݒ
	hr = m_pd3dDevice->SetRenderTarget(0, srf);
	srf->Release();

	return hr;
}

//sNZtH[}bg̎w
HRESULT GpuFilter::SetPixelFmt(PIXEL_FORMAT fmt)
{
	if(fmt < S16 || fmt >= Unknown) return E_FAIL;
	
	switch(fmt){
		case S16:
			if(m_pixFormatAvailable[fmt]){
				m_pixFormat = S16;
				break;
			}
		case U16:
		case Unknown:
		default:
			m_pixFormat = U16;
	}

	return S_OK;
}

//PIXEL_FORMATD3DFORMATɕϊ
D3DFORMAT GpuFilter::PicFmtToD3DFmt(PIXEL_FORMAT p) const
{
	switch(p){
		case S16:
			return D3DFMT_Q16W16V16U16;
		case U16:
			return D3DFMT_A16B16G16R16;
		case Unknown:
		default:
			return D3DFMT_UNKNOWN;
	}
}

//ʃXbhŏ
void __cdecl GpuFilter::AsyncProc(void *param)
{
	GpuFilter *pF = static_cast<GpuFilter*>(param);
	
	if(pF){
		if(pF->ProcImage()){
			pF->GetRenderTargetData();
		}
	}
}

#ifdef PROFILE_GPU
//fobOpNG
HRESULT GpuFilter::CreateQuery()
{
	SAFE_RELEASE(m_pQuery);
	return m_pd3dDevice->CreateQuery(D3DQUERYTYPE_EVENT, &m_pQuery);
}
#endif

//---------------------------------------------------------------------
// internal functions
//---------------------------------------------------------------------

//ϊRs[(line) PIXEL_YC -> D3DFMT_Q16W16V16U16 MMXp
void yc_i16_mmx(BYTE *dst, const PIXEL_YC *src, unsigned int n)
{
	unsigned int i;
	__m64 *mdst = (__m64*)dst;

	for(i=0; i<n; ++i){
		*mdst++ = *(__m64*)src++;
		//_mm_stream_pi(mdst++, *(__m64*)src++);
	}
}

//ϊRs[ D3DFMT_Q16W16V16U16 -> PIXEL_YC MMXp
void i16_yc_mmx(PIXEL_YC *dst, const BYTE *src, unsigned int n)
{
	unsigned int i;
	const __m64 *msrc = (const __m64*)src;

	for(i=0; i<n; ++i){
		*(__m64*)dst++ = *msrc++;
	}
}

//ϊRs[(line) PIXEL_YC -> D3DFMT_A16B16G16R16 MMXp
void yc_u16_mmx(BYTE *dst, const PIXEL_YC *src, unsigned int n)
{
	unsigned int i;
	__m64 *mdst = (__m64*)dst;
	const __m64 u16_ofs = { 0x8000800080008000UI64 };

	for(i=0; i<n; ++i){
		*mdst++ = _mm_add_pi16(*(__m64*)src++, u16_ofs);
		//_mm_stream_pi(mdst++, _mm_add_pi16(*(__m64*)src++, u16_ofs));
	}
}

//ϊRs[ D3DFMT_A16B16G16R16 -> PIXEL_YC MMXp
void u16_yc_mmx(PIXEL_YC *dst, const BYTE *src, unsigned int n)
{
	unsigned int i;
	const __m64 *msrc = (const __m64*)src;
	const __m64 u16_ofs = { 0x8000800080008000UI64 };

	for(i=0; i<n; ++i){
		*(__m64*)dst++ = _mm_sub_pi16(*msrc++, u16_ofs);
	}
}


static void __cdecl cp_yc_i16_mt(int id, int num, void *p0, void *p1)
{
	const MTCP_PARAM *para = static_cast<MTCP_PARAM*>(p0);
	long i;
	BYTE *dst = para->pSurf;
	PIXEL_YC *src = para->pYC;

	if(num){
		while((i = InterlockedExchangeAdd(&mt_cnt, 1)) < para->h){
			yc_i16_mmx(dst + i * para->sPitch, src + i * para->ycPitch, para->w);
		}
	}else{
		for(i=0; i<para->h; ++i){
			yc_i16_mmx(dst, src, para->w);
			dst += para->sPitch;
			src += para->ycPitch;
		}
	}

	_mm_empty();
}

static void __cdecl cp_i16_yc_mt(int id, int num, void *p0, void *p1)
{
	const MTCP_PARAM *para = static_cast<MTCP_PARAM*>(p0);
	long i;
	BYTE *src = para->pSurf;
	PIXEL_YC *dst = para->pYC;

	if(num){
		while((i = InterlockedExchangeAdd(&mt_cnt, 1)) < para->h){
			i16_yc_mmx(dst + i * para->ycPitch, src + i * para->sPitch, para->w);
		}
	}else{
		for(i=0; i<para->h; ++i){
			i16_yc_mmx(dst, src, para->w);
			dst += para->ycPitch;
			src += para->sPitch;
		}
	}

	_mm_empty();
}

static void __cdecl cp_yc_u16_mt(int id, int num, void *p0, void *p1)
{
	const MTCP_PARAM *para = static_cast<MTCP_PARAM*>(p0);
	long i;
	BYTE *dst = para->pSurf;
	PIXEL_YC *src = para->pYC;

	if(num){
		while((i = InterlockedExchangeAdd(&mt_cnt, 1)) < para->h){
			yc_u16_mmx(dst + i * para->sPitch, src + i * para->ycPitch, para->w);
		}
	}else{
		for(i=0; i<para->h; ++i){
			yc_u16_mmx(dst, src, para->w);
			dst += para->sPitch;
			src += para->ycPitch;
		}
	}

	_mm_empty();
}

static void __cdecl cp_u16_yc_mt(int id, int num, void *p0, void *p1)
{
	const MTCP_PARAM *para = static_cast<MTCP_PARAM*>(p0);
	long i;
	BYTE *src = para->pSurf;
	PIXEL_YC *dst = para->pYC;

	if(num){
		while((i = InterlockedExchangeAdd(&mt_cnt, 1)) < para->h){
			u16_yc_mmx(dst + i * para->ycPitch, src + i * para->sPitch, para->w);
		}
	}else{
		for(i=0; i<para->h; ++i){
			u16_yc_mmx(dst, src, para->w);
			dst += para->ycPitch;
			src += para->sPitch;
		}
	}

	_mm_empty();
}
